/* Rick's little internet radio

   For the ESP32. This is a no-bull version. It will connect, play a station
   (supporting redirection), and handle displaying metadata on an LCD for
   title/artist.

   Version 0.09  15th January 2021

   https://heyrick.eu/blog/index.php?diary=20210115

   Licenced under the EUPL (v1.1 only).

   NVRAM use:
      (str)  rnr.ssid      - WiFi access point name
      (str)  rnr.pass      - WiFi password
      (uint) rnr.vol       - Volume to use at start
      (unit) rnr.cstn      - Current station to use at start
      (str)  rnr.stn#name  - Name of station #
      (str)  rnr.stn#url   - URL if station #
                             (# is from 0 to 9)
*/


// Do we want debug information?
// #define DEBUG 1


// Include some standard libraries
#include <WiFi.h>                   // WiFi support
#include <HTTPClient.h>             // Fetch stuff from the internet
#include <esp_wifi.h>               // ESP32 specific WiFi functions
#include <Wire.h>                   // IIC
#include <Preferences.h>            // ESP32 settings in NVRAM
#include <Ticker.h>                 // Ticker event


// Now include additional libraries to work with the hardware we're using
//   VS1053 is from https://github.com/baldram/ESP_VS1053_Library
//     download it and install it from zip file
//   LiquidCrystal_PCF8574 is available in the library manager
//     it's the one by Matthias Hertel
#include <VS1053.h>                 // VS0153 based MP3 decoder board
#include <LiquidCrystal_PCF8574.h>  // IIC connected 1602 LCD


// We're using the standard SPI pins, but we will need to define the
// extra pins for CS, DCS, and DREQ.
#define VS1053_CS    32
#define VS1053_DCS   33
#define VS1053_DREQ  35


// The volume level goes from 0 (off) to 100 (max).
// This volume provides a reasonable level to give me music without
// disturbing others when plugged into speakers beside my monitor
#define INITVOLUME   85


// The 1602 LCD is controlled by a PCF8574 with a default base address of &27
#define IICADDR      0x27


// Buttons
//    Button           Short press    Long press
//      + (next)       Vol+           Next station
//      - (prev)       Vol-           Previous station
#define NEXTBUTTON   13
#define PREVBUTTON   12

#define ACTIONNONE   0
#define ACTIONPRESS  1
#define ACTIONLONGPRESS 2


// Definitions for changing the volume/station
#define VOLUP        1
#define VOLDOWN      0
#define NEXT         1
#define PREV         0


// Storage for the AP's SSID and password.
char wifissid[64] = "";
char wifipass[64] = "";


// Storage for the station URL (in case of redirects)
char stnurl[256] = "";


// Project ID (must be <16 chars for LCD)
char appname[] = "Rick's NetRadio\xBC"; // &BC is "shi" in Katakana, like a little smiley :)
char hostname[] = "RicksNetRadio";      // must be UNIQUE if you have more than one radio...


// Version/URL strings
char appvers[] = "0.09";
char appurl[] = "https://heyrick.eu/blog/index.php?diary=20210115";


// Define radio stations
typedef struct stationdef
{
   char name[32];         // Textual name of station ("Eagle '80s")
   char url[128];         // Full station URL
};

struct stationdef station[10] =
{
   // These stations are used as the default if there are no settings to provide custom stations.
   //

   // First station - this is the station played at startup
   "Heart 80s",
   "http://media-ice.musicradio.com/Heart80s",

   // Second station
   "Love 80s (M)",
   "http://rss-streaming.co.uk:8042/;",

   // Third station
   "Classic FM",
   "http://media-ice.musicradio.com/ClassicFM",

   // Fourth station
   "Retro Hits Canada",
   "http://149.56.183.178:8300/stream",

   // Fifth station
   "BBC Radio 4 FM",
   "http://bbcmedia.ic.llnwd.net/stream/bbcmedia_radio4fm_mf_p",

   // Sixth station
   "Alouette",
   "http://alouette.ice.infomaniak.ch/alouette-high.mp3",

   //// Fifth station
   //"RTE1",      <-- whatever this does on connect, it crashes the ESP!
   //"av.rasset.ie",
   //"/av/live/radio/adio1.m3u",
   //80,

   // Seventh station
   "JPopsuki Radio!",
   "http://jpopsuki.fm:8000/stream",

   // Eighth station
   "Birdsong FM",
   "http://streaming.radio.co/s5c5da6a36/listen",

   // Ninth station
   "Gothique13",
   "http://66.55.145.43:7037/",

   // Tenth station (here, so it can be accessed quickly by going back once)
   "PPN Radio",
   "http://streamingV2.shoutcast.com/PPNR02"
};

struct buttondef
{
   bool state;
   unsigned long press;
   unsigned long release;
   int  action;
} button[2];


// The globals
int  lcdpresent = 0;      // Set to '1' if LCD is detected
int  connected = 0;       // Set to '1' if connected to a station
int  bitrate = 0;         // The stream bitrate (in kbit)
int  metaint = 0;         // The interval of when metadata appears (usually 8192-32768)
int  bytecount = 0;       // The number of bytes to go until there's a metadata block
char metadata[8192];      // ICY metadata - global for speed (can be up to 4080 bytes)
int  currentstation = 0;  // Currently selected station
int  stationcount = 10;   // How many stations we have defined (keep this fixed at 10 now)
int  volume = INITVOLUME; // Current volume
char title[32] = "";      // Current song title
char ptitle[32] = "";     // Previous song title
char artist[32] = "";     // Current song artist
char serinput[64] = "";   // Serial input string
int  wantmetadata = 1;    // Do we want metadata?
unsigned long lcddelay=0; // Delay for temporary messages on LCD (0=none, also is a timeout)
unsigned long totaldata = 0; // Watchdog delay


// Define an instance of the 16x2 LCD
LiquidCrystal_PCF8574 lcd(IICADDR);


// Define an instance of the VS1053, and give it an aligned byte buffer
// Buffer size was 32 bytes, but 64 seems to work well too, less churning.
#define BUFSIZE 64
VS1053 audioplayer(VS1053_CS, VS1053_DCS, VS1053_DREQ);
__attribute__((aligned(4))) uint8_t buffer[BUFSIZE]; // word aligned for speed


// Define an instance of the WiFi client
WiFiClient webclient;


// Define an instance of the NVRAM settings
Preferences settings;


// Define an instance of the icker
Ticker ticker;



// *************************
// *                       *
// * DEVICE INITIALISATION *
// *                       *
// *************************

void setup ()
{
   char serinfo[64];
   int  result = 0;

   // Set up serial port for tracing activity (debugging) and/or control and setup
   Serial.begin(115200);
#ifdef DEBUG
   Serial.print(appname);
   Serial.println(" starting up.");
   sprintf(serinfo, "Running on CPU %d at %dMHz, with %d memory free.",
           xPortGetCoreID(), ESP.getCpuFreqMHz(), ESP.getFreeHeap());
   Serial.println(serinfo);
   sprintf(serinfo, "(version %s built at %s on %s)", appvers, __TIME__, __DATE__);
   Serial.println(serinfo);
   Serial.println("");
#else
   Serial.println("");
   Serial.println("Rick's NetRadio is initialising...");
   // always report the version and build date
   sprintf(serinfo, "(version %s built at %s on %s)", appvers, __TIME__, __DATE__);
   Serial.println(serinfo);
   Serial.println("");
#endif

   // Is an LCD connected?
#ifdef DEBUG
   Serial.print("Looking for LCD...");
#endif
   Wire.begin();
   Wire.beginTransmission(IICADDR);
   if ( Wire.endTransmission() == 0 )
   {
#ifdef DEBUG
      Serial.println("found.");
#endif
      lcdpresent = 1;

      // Initialise the LCD (we can't assume anything about its state)
      lcd.begin(16, 2);
      lcd.setBacklight(127);
      lcd.home();
      lcd.clear();
      lcd.noBlink();
      lcd.noCursor();
      lcd.setCursor(0, 0);
      lcd.print(appname);
      lcd.setCursor(0, 1);
      lcd.print("initialising...");
   }
   else
   {
      // Always write something to serial if device not detected
      Serial.println("16x2 LCD was not detected (is IICADDR correct?).");
   }

   // Set up the blue LED (will light when connected to a station)
   pinMode(2, OUTPUT);
   digitalWrite(2, LOW); // default to off

   // Set up the previous/next buttons
   pinMode(NEXTBUTTON, INPUT_PULLUP);  // pulled up, so ground
   pinMode(PREVBUTTON, INPUT_PULLUP);  // to activate the button
   button[0].state = HIGH;
   button[1].state = HIGH;
   button[0].action = ACTIONNONE;
   button[1].action = ACTIONNONE;

   // Set up SPI
   SPI.begin();

   // Initialise the MP3 decoder - this takes a moment
   mp3_initialise();

   // Initialise the settings
   settings.begin("rnr", false); // "rnr" = Rick's Net Radio

   // Do we have WiFi information? If not, toss us into presetup.
   strncpy(wifissid, settings.getString("ssid", "").c_str(), 63);
   wifissid[63] = '\0';
   if ( strlen(wifissid) == 0 )
   {
      Serial.println("No WiFi access point name read from NVRAM, entering setup.");
      presetup();
   }
   strncpy(wifipass, settings.getString("pass", "").c_str(), 63);
   wifipass[63] = '\0';
   // Don't fault the password being blank, maybe there is no password...

   // Load the stations.
   if ( savedstation_count() > 0 )
   {
#ifdef DEBUG
      Serial.println("Saved stations found, loading them.");
#endif
      savedstation_loadall(); // Stations are defined, so load them in
   }
   else
   {
#ifdef DEBUG
      Serial.println("No stations saved, so setting defaults.");
#endif
      savedstation_saveall(); // Nothing defined, so save defaults (so they can be edited)
   }

   // Now connect to the AP
   wifi_connect();

   // Set up the watchdog
   ticker.attach(10.0, watchdog);

   // Read the previously selected volume and station
   volume = settings.getUInt("vol", INITVOLUME); // default to initial setting
   currentstation = settings.getUInt("cstn", 0);

   // Verify that currentstation points to an actual station.
   while ( strlen(station[currentstation].url) == 0 )
   {
      sprintf(serinfo, "Station %d invalid, trying next...", currentstation);
      Serial.println(serinfo);

      currentstation++;
      if ( currentstation == stationcount )
         currentstation = 0;
   }

   // Done initialising, we'll resume in loop...
}



// *************
// *           *
// * MAIN LOOP *
// *           *
// *************

void loop()
{
   // Round and round we go...
   //
   // Warning: Adding more to this loop will probably require
   //          some sort of buffering to be implemented...

   uint8_t bytes = 0;
   int  reply = 0;

   // Connect?
   if ( !connected )
   {
      // Keep trying until connected (allows us to handle redirections)
      do
      {
         reply = station_connect();
         check_buttons(); // so user can choose another station
         check_serial();  // or fix the problem
      } while ( reply == 0 );

      // Always report when connected to a new station
      Serial.println("");
      Serial.print("Connected to ");
      Serial.println(station[currentstation].name);
      Serial.println("");
      Serial.println("Controls: [P]revious/[N]ext station, [L]ouder/[Q]uieter, [M]ute,");
      Serial.println("          show [A]ll stations, [0]-[9] direct station choice,");
      Serial.println("          [S]etup (playing stops), [C]opyright, or [R]eset device.");
      Serial.println("");

      // How many bytes to go until we need to deal with metadata
      bytecount = metaint;

      // Prior to playing, set the volume
      audioplayer.setVolume(volume);
   }

   // If there's data, handle it in BUFSIZE byte chunks, catering for metadata
   if (webclient.available() > 0)
   {
      // Read BUFSIZE bytes and throw them directly to the MP3 player
      // (this actualy works as long as nothing too complex happens in loop()!)

      if ( metaint != 0 )
      {
         // We have metadata, so we need to read the appropriate amounts

         if ( bytecount >= BUFSIZE )
         {
            // It's okay, we have data to go until metadata block
            bytes = webclient.read(buffer, BUFSIZE);
            audioplayer.playChunk(buffer, bytes);

            totaldata += bytes;

            // Subtract bytes from the bytecount
            bytecount -= bytes;
            if (bytecount == 0)
            {
               // If it's reached zero, check to see if there is information to read
               read_icy_metadata();
               bytecount = metaint; // reset byte count for next bit of metadata
            }
         }
         else
         {
            // We have less than BUFSIZE bytes before metadata, so read what's left
            bytes = webclient.read(buffer, bytecount);
            audioplayer.playChunk(buffer, bytes);

            totaldata += bytes;

            // Subtract bytes from the bytecount (should equal zero!)
            bytecount -= bytes;
            if (bytecount == 0)
            {
               // If it's reached zero, check to see if there is information to read
               read_icy_metadata();
               bytecount = metaint; // reset byte count for next bit of metadata
            }
         }
      }
      else
      {
         // There is no metadata, so just read and play
         bytes = webclient.read(buffer, BUFSIZE);
         audioplayer.playChunk(buffer, bytes);
         totaldata += bytes;
      }
   }

   // Check the buttons
   check_buttons();

   // Serial?
   check_serial();

   // LCD to twiddle?
   if ( lcddelay != 0 )
   {
      // Expired?
      if ( lcddelay < millis() )
      {
         // Yes, so restore what should normally be shown
         lcd_title();
         lcddelay = 0;
      }
   }

   // Warning: Adding more to this loop will probably require
   //          some sort of buffering to be implemented...
}



// ***********************
// *                     *
// * CONNECTION WATCHDOG *
// *                     *
// ***********************

void watchdog(void)
{
   // Called every TEN seconds.If we're connected and the data count hasn't changed,
   // then something has gone wrong and we should reconnect to this station.

   static unsigned long oldtotaldata = 12345;

   if ( connected )
   {
      if ( totaldata == oldtotaldata )
      {
         Serial.println("*** WATCHDOG FIRED - reconnecting to station.");
         connected = 0;
      }

      oldtotaldata = totaldata;
   }
}



// ****************
// *              *
// * LCD ROUTINES *
// *              *
// ****************

void lcd_output(char *lineone, char *linetwo)
{
   // Write the two lines given to the LCD

   if ( !lcdpresent )
      return;

   lcd.clear();
   lcd.setCursor(0, 0);
   lcd.print(lineone);
   lcd.setCursor(0, 1);
   lcd.print(linetwo);

   return;
}



void lcd_title()
{
   // Output something on the LCD to reflect title, or failing
   // that, the station.
   char report[32] = "";

   if ( strlen(title) == 0 )
   {
      // We have no title

      if ( strlen(artist) == 0 )
      {
         // We have no artist either, so output station info
         sprintf(report, "%dkbit MP3", bitrate);
         lcd_output(station[currentstation].name, report);
      }
      else
      {
         // We have an artist but no title (probably omething odd in StreamInfo)
         // so display station name and whatever is in artist.
         lcd_output(station[currentstation].name, artist);
      }
   }
   else
   {
      // We have a title (and thus presumably an artist) so display it
      lcd_output(title, artist);
   }
}



// *************************
// *                       *
// * START UP MP3 HARDWARE *
// *                       *
// *************************

void mp3_initialise()
{
   // Initialise the VS1053 to play in MP3 mode
   audioplayer.begin();
   if ( audioplayer.isChipConnected() == 0 )
   {
      lcd_output("VS1053 audio IC", "not responding");
      delay(2500); // nothing much we can do here...
   }
   audioplayer.switchToMp3Mode();
   audioplayer.setVolume(volume);
}



// ***************************
// *                         *
// * CONNECT TO ACCESS POINT *
// *                         *
// ***************************

void wifi_connect()
{
   // Try to connect to the AP
   //
   // My board always fails to connect the first time,
   // so we'll run a timeout that if it doesn't
   // manage to connect in three and a half seconds,
   // do a hard reset. This works. Dunno why...
   //

   int  timeout = 0;
   long sigstrength = 0;
   char sigreport[32] = "";

   lcd_output("WiFi connecting:", wifissid);
#ifdef DEBUG
   Serial.print("Trying to connect to ");
   Serial.println(wifissid);
#endif

   // Kill off any previous behaviour
   WiFi.disconnect(true);

   // Start the WiFi system
   WiFi.mode(WIFI_STA);
   WiFi.begin(wifissid, wifipass);
   WiFi.setHostname(hostname);

   // Wait while it connects
   while (WiFi.status() != WL_CONNECTED)
   {
      delay(350);
#ifdef DEBUG
      Serial.print(".");
#endif

      // Allow serial to break in if we're stuck with an AP that no longer exists
      check_serial();

      // If it's taken too long, force a restart
      // (this usually works - no idea why)
      timeout += 1;
      if ( timeout > 10 )
      {
         Serial.println("WiFi connection timeout, restarting...");
         lcd_output("Restarting...", "");
         delay(500);
         ESP.restart();
      }
   }

   // Report connected, and WiFi signal strength
   sigstrength = WiFi.RSSI();
#ifdef DEBUG
   sprintf(sigreport, "Connect %lddBm", sigstrength);
   Serial.println(sigreport);
#endif
   lcd_output(sigreport, wifissid);
   delay(1500); // delay so message can be seen
}



// ************************
// *                      *
// * CONNECT TO A STATION *
// *                      *
// ************************

int station_connect()
{
   // Connect to the current station, returns ZERO if did NOT connect.
   // Try again - host/path may have been updated if redirect.
   //
   // NOTE: If stnurl is EMPTY then we will read the URL from the current
   //       station definition.
   //       If stnurl is NOT empty, then we will use that, as this may
   //       be some sort of redirection.
   //

   String header;
   String newhost;
   char hdrbuf[512] = "";
   char host[128] = "";
   char path[128] = "";
   int  port = 80; // unless specified otherwise
   char *work;
   int  hdrposn = 0;
   int  twonewlines = 0;
   int  thisbyte = 0;
   int  lastbyte = 0;

   digitalWrite(2, LOW); // turn off blue LED
   lcd_output("Connecting to", station[currentstation].name);

   // If the workspace URL is empty, then copy in the station's URL
   // If it is NOT empty, we came here through a redirect.
   if ( strlen(stnurl) == 0 )
      strcpy(stnurl, station[currentstation].url);

   Serial.print("Connecting to ");
   Serial.println(stnurl);

   // Now split the URL into parts
   if ( strstr(stnurl, "http://") == stnurl )
      work = stnurl + 7;         // skip "http://"
   else
      work = stnurl;             // user is an idiot and didn't write a proper URL, assume hat we have is a hostname

   strcpy(host, work);           // host contains entire URL without prefix
   work = strchr(host, '/');     // look for path separator
   if ( work == NULL )
   {
      strcpy(path, "/");         // user is an idiot and didn't write a proper URL, assume path is '/'
   }
   else
   {
      strcpy(path, work);        // extract the path
      work[0] = '\0';            // truncate the host
   }

   work = strchr(host, ':');     // look for port seperator
   if ( work != NULL )
   {
      work[0] = '\0';            // truncate the host
      work++;                    // advance to next charactor
      port = atoi(work);         // extract the port number
   }

#ifdef DEBUG
   Serial.println("URL split to: ");
   Serial.print("  Host: \"");
   Serial.print(host);
   Serial.println("\"");
   Serial.print("  Path: \"");
   Serial.print(path);
   Serial.println("\"");
   Serial.print("  Port: ");
   Serial.println(port);
#endif

   title[0] = '\0';
   artist[0] = '\0';
   metaint = 0;

   // Open a connection
   if ( webclient.connect(host, port) )
   {
#ifdef DEBUG
      Serial.println("Port opened, sending request.");
#endif

      if ( wantmetadata )
      {
         // We want Icy metadata
         webclient.print(String("GET ") + path + " HTTP/1.1\r\n" +
                         "Host: " + host + "\r\n" +
                         "Icy-MetaData:1\r\n" +
                         "Connection: close\r\n\r\n");
      }
      else
      {
         // We do NOT want Icy metadata
         webclient.print(String("GET ") + path + " HTTP/1.1\r\n" +
                         "Host: " + host + "\r\n" +
                         "Connection: close\r\n\r\n");
      }

      // Now we need to read header lines and extract data
      delay(500); // wait for some data to come back
      if (webclient.available() == 0)
      {
         lcd_output(station[currentstation].name, "No data,retrying");
         delay(1500); // 1.5 sec delay before returning failed
         strcpy(stnurl, ""); // blank the parsed URL
         return 0;
      }

      // Keep reading until we read two LF bytes in a row
      while ( !twonewlines )
      {
         // Read a line
         hdrposn = 0;
         hdrbuf[0] = '\0';
         do
         {
            // Read a byte
            thisbyte = webclient.read();
            if ( thisbyte > 31 )
            {
               // If a printable, add it to the buffer
               hdrbuf[hdrposn++] = thisbyte;
               hdrbuf[hdrposn] = '\0';
               lastbyte = thisbyte;

               // If too long, just loop back. It's probably just
               // overlong rubbish like a cookie line or somesuch...
               if ( hdrposn > 511 )
                  hdrposn = 0;
            }
            else
            {
               // It's a control character - is it an LF?
               if ( thisbyte == 10 )
               {
                  // If this byte is LF and so was the last one...
                  if ( lastbyte == thisbyte )
                     twonewlines = 1; // flag we've reached the end of the headers

                  lastbyte = thisbyte;
               }
            }
         } while ( thisbyte != 10 );

         // We have a header line, so let's work out what it is
#ifdef DEBUG
         Serial.println(hdrbuf);
#endif
         header = hdrbuf;

         // Deal with "Location:" header for redirection
         header.toLowerCase();
         if ( header.startsWith("location: http://") )
         {
            // Work out where we're supposed to point now
            work = hdrbuf + 10; // skip "location: "

            // Copy it to the stnurl variable
            strcpy(stnurl, work);

#ifdef DEBUG
            Serial.print("Redirection -> ");
            Serial.println(stnurl);
#endif

            // Don't try closing the connection, it'll already have been done
            // and doing it ourselves will cause the device to hang (great code!)

            lcd_output("Redirecting...", "");
            delay(500); // half a second so message can be seen (briefly!)

            return 0;
         }

         // Read proper name of station
         // look for "icy-name:" and take substring(9)

         // Read station bitrate
         if ( header.startsWith("icy-br:" ) )
            bitrate = header.substring(7).toInt();

         // Read ICY info interval
         if ( header.startsWith("icy-metaint:" ) )
            metaint = header.substring(12).toInt();
      }

      // If we have a bitrate, write that to the LCD in line two
      if ( bitrate > 0 )
      {
         lcd_title(); // will put bitrate into second line as no title/artist info yet
      }
      else
      {
         // No bitrate given, so output the hostname in the second line
         lcd_output(station[currentstation].name, host);
      }

#ifdef DEBUG
      sprintf(hdrbuf, "Connection to", station[currentstation].name);
      Serial.println(hdrbuf);
#endif
   }
   else
   {
#ifdef DEBUG
      Serial.println("Connection failed.");
#endif
      lcd_output(station[currentstation].name, "Connect failed!");
      delay(1500);

      strcpy(stnurl, "");

      return 0;
   }

   // Assume by now that we're connected
   connected = 1;
   digitalWrite(2, HIGH); // turn on blue LED

   strcpy(stnurl, "");

   return 1; // connected!
}



// *******************************
// *                             *
// * READ AND PARSE ICY METADATA *
// *                             *
// *******************************

void read_icy_metadata()
{
   // Read the metadata, look for a title

   int  mdatlen = 0;
   int  recurse = 0;
   int  corrupt = 0;
   char *start = NULL;
   char *end = NULL;

   // Read length byte
   mdatlen = webclient.read();

#ifdef DEBUG
   Serial.println("Reading metadata");
#endif

   // No data?
   if ( mdatlen == 0 )
      return;

   // The size is actually multiplied by 16, to allow up to 4080 bytes from a single size byte
   // (the payload size does not include the size byte)
   mdatlen = mdatlen * 16;

   // Read the data
   for ( recurse = 0; recurse < mdatlen; recurse++ )
   {
      metadata[recurse] = webclient.read();

      // I don't know if TAB is valid in metadata, but bytes 1-8 aren't, so if
      // we see any of those, assume the data has been corrupted and that we
      // can no longer look for metadata as we're out of sync.
      // (a poor WiFi signal can cause data loss which will trigger this)
      if ( (metadata[recurse] > 0) && (metadata[recurse] < 9) )
         corrupt = 1;
   }
   metadata[mdatlen] = '\0';
#ifdef DEBUG
   Serial.print("Metadata \"");
   Serial.print(metadata);
   Serial.println("\"");
#endif

   // If corrupt, note this and abort
   if ( corrupt )
   {
      // There's no possible recovery, metadata does not have any distinctive
      // header, so just flag us as not connected so we can reconnect.
      lcd_output(station[currentstation].name, "Metadata SyncErr");
      connected = 0;
      digitalWrite(2, LOW); // turn off blue LED

      return;
   }

   // Okay, now let's look for the title and artist names
   // We will see something like:
   //   StreamTitle='Kirsty MacColl - Days';
   // or:
   //   StreamTitle=' - ';
   // The latter being in between songs, during adverts, etc.
   // There may be other metadata, such as "StreamUrl='';".
   // May also be:
   //   StreamTitle='';
   // if the channel doesn't bother with information
   // (but has to support it as the client will be expecting it)

   start = strstr(metadata, "StreamTitle");
   if ( start != NULL )
   {
      // We have found the streamtitle
      start += 12; // skip over "StreamTitle="

      // Now look for the end marker
      end = strstr(start, ";");
      if ( end == NULL )
      {
         // No end, so just set it as the string length
         end = (char *)start + ( strlen(start) - 1);
      }

      // Quoted string?
      if ( start[0] == '\'')
      {
         // It seems as if quotes may be optional, so handle them.
         start += 1;
         end -= 1;
      }

      // Terminull the part we want
      end[0] = '\0';

      // Now start should point to a string like:
      //   "Bananarama - Venus"
      // Or maybe just " - ", or even just "".

      // Trap empty metadata
      if ( strcmp(start, " - ") == 0 )
      {
         // No metadata right now, so display the station name
         title[0] = '\0';
         artist[0] = '\0';
         lcd_title();
         return;
      }

      // Okay, sort out what's the title and what's the artist
      end = strstr(start, " - ");
      if ( end == NULL )
      {
         // Separator not found, output station title and this string
         title[0] = '\0';
         strncpy(artist, start, 16);
         artist[16] = '\0';
         lcd_title();
      }
      else
      {
         // First part is artist, second part is title
         end[0] = '\0'; // Terminate artist part
         end += 3; // Skip to start of title
         strncpy(title, end, 16);
         title[16] = '\0';
         strncpy(artist, start, 16);
         artist[16] = '\0';
         lcd_title();

         // Has the title changed?
         if ( strcmp(title, ptitle) != 0 )
         {
            // Different
            if ( strlen(title) > 0 )
            {
               // And title is non-zero, so set this as the new title
               strcpy(ptitle, title);
               // And print it out to serial
               Serial.print("Currently playing \"");
               Serial.print(end); // title, unclipped version
               Serial.print("\" by ");
               Serial.print(start); // artist, unclipped version
               Serial.println(".");
            }
         }

         // ##TODO##
         // As I write this JPopsuki has just given me:
         //   StreamTitle='キャプテンストライダム - バースデー';
         // Which on the LCD looks like:
         //   ɛ  ɛ シɛ ケɛ  ɛ シ
         //   ɛ ュɛ 」ɛ  ɛ  ɛ ウɛ
         // So I think in the future it might be useful
         // to detect UTF-8 and copy across as '?' or
         // something for the things that can't be shown.
         //
         // LCD character set (Hitachi HD44780):
         //   !"#$%&'()*+,-./
         //  0123456789:;<=>?     <-- this is like ASCII
         //  @ABCDEFGHIJKLMNO
         //  PQRSTUVWXYZ[¥]^_     <-- but note the '¥' instead of '\'
         //  `abcdefghijklmno
         //  pqrstuvwxyz{|}←→    <-- arrows instead of ~ and unprintable
         //  [ not defined ]
         //  [ not defined ]
         //   。「」、・ヲァィゥェォャュョッ     <-- small katakana
         //  ㆒アイウエオカキクケコサシスセソ     <-- katakana
         //  タチツテトナニヌネノハヒフヘホマ
         //  ミムメモヤユヨラリルレロワン゛゜
         //  αäßɛµσρℊ√₁j̽¢Lñö     <-- more or less, greek/maths
         //  ρϕθ∞ΩüΣπ※У千万円÷ ▊     <-- more or less, greek/maths
         //
         // (if that looked like gibberish, your editor isn't in a UTF-8 mode!)
      }
   }
}



// ****************************
// *                          *
// * CHANGE VOLUME OR STATION *
// *                          *
// ****************************

void change_volume(int louder)
{
   // Handle changing the volume
   char report[32];
   char volbar[17];

   if ( louder )
   {
      // Increase volume
      volume += 5;
      if ( volume > 100 )
         volume = 100; // clip at 100
   }
   else
   {
      // Decrease volume
      volume -= 5;
      if ( volume < 5 )
         volume = 5;
   }
   audioplayer.setVolume(volume);
   settings.putUInt("vol", volume);

   // Construct the LCD report with volume bar
   sprintf(report, "Volume %d", volume);
#ifdef DEBUG
   Serial.println(report);
#endif
   for (int recurse = 0; recurse < 16; recurse++)
      volbar[recurse] = ' ';
   volbar[16] = '\0';
   for (int recurse = 0; recurse < (volume / 6.25); recurse++)
      volbar[recurse] = 255;
   volbar[16] = '\0';
   lcd_output(report, volbar);
   lcddelay = millis() + 1500; // go away after one and a half seconds

   sprintf(report, "The volume is now %d%%.", volume);
   Serial.println(report);

   return;
}



void change_station(int next)
{
   // Change station

    int   changeticks = 0;
    char  report[64] = "";

    do
    {
       if ( next )
       {
          currentstation += 1;
          if ( currentstation == stationcount )
             currentstation = 0; // wrap around
       }
       else
       {
          currentstation -= 1;
          if ( currentstation < 0 )
             currentstation = (stationcount - 1); // wrap around inverse
       }
       changeticks++;
    } while (  ( strlen(station[currentstation].url) == 0 ) && ( changeticks < 11 ) );

    if ( changeticks == 11 )
    {
       // Wrapped around without finding a suitable station, toss us into setup.
       presetup();
       return;
    }

    // Change to the selected channel
    sprintf(report, "Station change to %s (%d).", station[currentstation].name, currentstation);
    Serial.println(report);
    connected = 0; // to force connecting to new station
    digitalWrite(2, LOW); // turn off blue LED
    settings.putUInt("cstn", currentstation);

    return;
}



// **********************************
// *                                *
// * HANDLE HARDWARE BUTTON PRESSES *
// *                                *
// **********************************

void check_buttons()
{
   // Check the buttons and perform actions on button release (unless BOTH pressed)

   bool level = 0;
   unsigned long ticker = 0;
   int  thisbutton = 0;
   char report[32];
   char volbar[17];
   long sigstrength = 0;

   // Deal with the buttons
   for ( thisbutton = 0; thisbutton < 2; thisbutton++ )
   {
      level = (digitalRead( (PREVBUTTON + thisbutton) ) == HIGH );
      if ( level != button[thisbutton].state )
      {
         if ( !level )
         {
            // HIGH to LOW - button press
            button[thisbutton].press = millis();
            button[thisbutton].state = level;

            // Detect if the other button has been pressed as well
            if ( ( button[thisbutton].state == LOW ) &&
                 ( button[(thisbutton ^ 1)].state == LOW ) )
            {
               // Both buttons pressed - toggle whether or not we want metadata
               wantmetadata = wantmetadata ^ 1;
               sigstrength = WiFi.RSSI();
               sprintf(report, "WiFi %lddBm", sigstrength);
               if ( wantmetadata == 1 )
                  lcd_output("MetadataEnabled", report);
               else
                  lcd_output("MetadataDisabled", report);
               connected = 0;
               digitalWrite(2, LOW); // turn off blue LED
               button[0].state = HIGH;
               button[1].state = HIGH;
               delay(1000);
               return;
            }
         }
         else
         {
           // LOW to HIGH - button release
           button[thisbutton].release = millis();
           button[thisbutton].state = level;

           // Is there an action to perform?
           ticker = button[thisbutton].release - button[thisbutton].press;

           if ( ticker > 2500 )
           {
              // Pressed for longer than two and a half seconds, display some info
              if ( thisbutton == 1)
              {
                 // +ve button, report WiFi status
                 sigstrength = WiFi.RSSI();
                 sprintf(report, "WiFi is %lddBm", sigstrength);
                 if (   sigstrength >  -30 )                           strcpy(volbar, "(perfect)");
                 if ( ( sigstrength >= -50 ) && ( sigstrength < -30) ) strcpy(volbar, "(excellent)");
                 if ( ( sigstrength >= -60 ) && ( sigstrength < -50) ) strcpy(volbar, "(good)");
                 if ( ( sigstrength >= -70 ) && ( sigstrength < -60) ) strcpy(volbar, "(acceptable)");
                 if ( ( sigstrength >= -80 ) && ( sigstrength < -70) ) strcpy(volbar, "(poor)");
                 if ( ( sigstrength >= -90 ) && ( sigstrength < -80) ) strcpy(volbar, "(very poor)");
                 lcd_output(report, volbar);
                 lcddelay = millis() + 3000; // wait for longer before going away (3 sec)
              }
              else
              {
                 // -ve button, report system status
                 // 1234567890123456
                 // [no longer -> Core x @ 123MHz]
                 // MemFree: 123456
                 // Rx: 12345678
                 //sprintf(report, "Core %d @ %dMHz", xPortGetCoreID(), ESP.getCpuFreqMHz());
                 sprintf(volbar, "MemFree: %d", ESP.getFreeHeap());
                 sprintf(report, "Rx: %lu", totaldata);
                 lcd_output(volbar, report); // was report,volbar
                 lcddelay = millis() + 3000; // wait for longer before going away (3 sec)
              }
              return;
           }

           if ( ticker > 500 )
           {
              // Pressed for longer than half a second, it is a long press
              if ( thisbutton == 1 )
                 change_station(NEXT);
              else
                 change_station(PREV);
           }
           else
           {
              if ( ticker > 50 )
              {
                 // Held for more than 5cs, it is a regular press
                 if ( thisbutton == 1 )
                    change_volume(VOLUP);
                 else
                    change_volume(VOLDOWN);

              } // if ticker > 50 block
           } // if ticker > 500 block
         } // high to low or low to high block
      } // if level different to button block
   } // for thisbutton ... loop
}



// ****************************
// *                          *
// * HANDLE SERIAL PORT INPUT *
// *                          *
// ****************************

void check_serial(void)
{
   // Check to see if there is any serial input.
   //
   // We support (in alphabetical order):
   //   [0] - [9] to directly choose a station
   //   [A]ll stations (list)
   //   [B]ytecount (undocumented)
   //   [C]opyright
   //   [H] (and [?]) to give keypress help
   //   [L]ouder
   //   [M]ute
   //   [N]ext station
   //   [P]revious station
   //   [Q]uieter
   //   [R]eset
   //   [S]etup
   //   [W]ifi AP list (undocumented)
   //

   int  inbyte = 0;
   int  recurse = 0;
   char report[128];

   if ( Serial.available() > 0 )
   {
      inbyte = Serial.read();

      switch (inbyte)
      {
         case '0' : // Directly select station
         case '1' :
         case '2' :
         case '3' :
         case '4' :
         case '5' :
         case '6' :
         case '7' :
         case '8' :
         case '9' : recurse = inbyte - '0'; // convert ASCII to number
                    if ( strlen(station[recurse].url) == 0 )
                    {
                       // no URL, invalid
                       Serial.println("Undefined station, please choose another.");
                    }
                    else
                    {
                       // directly choose this station
                       currentstation = recurse;
                       connected = 0; // to force connecting to new station
                       digitalWrite(2, LOW); // turn off blue LED
                       settings.putUInt("cstn", currentstation);
                       sprintf(report, "Changing to %s.", station[currentstation].name);
                       Serial.println(report);
                    }
                    break;

         case 'A' : // All available stations
         case 'a' : Serial.println("");
                    Serial.println("Available stations:");
                    for (recurse = 0; recurse < 10; recurse++)
                    {
                       if ( strlen( station[recurse].url) == 0 )
                          sprintf(report, " %d: undefined", recurse);
                       else
                          sprintf(report, " %d: %s (%s)", recurse, station[recurse].name, station[recurse].url);
                       Serial.println(report);
                    }
                    Serial.println("");
                    break;

         case 'B' : // Bytecount (undocumented)
         case 'b' : Serial.println("");
                    Serial.print(totaldata);
                    Serial.println(" bytes have been received since the last reset.");
                    Serial.print("Thie current station is sending data at ");
                    Serial.print(bitrate);
                    Serial.println("kbps.");
                    Serial.println("");
                    break;

         case 'C' : // Copyright (don't alter this, it's my only claim to fame)
         case 'c' : Serial.println("");
                    Serial.println("****************************************************");
                    Serial.print("*          Rick's NetRadio, version ");
                    Serial.print(appvers);
                    Serial.println(".          *");
                    Serial.println("*        Copyright (C) 2021  Richard Murray        *");
                    Serial.print("* ");
                    Serial.print(appurl);
                    Serial.println(" *");
                    Serial.println("*    This project is licenced EUPL v1.1 (only).    *");
                    Serial.println("****************************************************");
                    Serial.println("");
                    break;

         case 'H' : // Help
         case 'h' :
         case '?' : Serial.println("");
                    Serial.println("Controls: [P]revious/[N]ext station, [L]ouder/[Q]uieter, [M]ute,");
                    Serial.println("          show [A]ll stations, [0]-[9] direct station choice,");
                    Serial.println("          [S]etup (playing stops), [C]opyright, or [R]eset device.");
                    Serial.println("");
                    break;

         case 'L' : // Louder
         case 'l' : change_volume(VOLUP);
                    break;

         case 'M' : // Mute
         case 'm' : if ( audioplayer.getVolume() == 0 )
                    {
                       // Audio was muted, unmute it
                       audioplayer.setVolume(volume);
                       lcd_output("     Audio", "    Unmuted");
                       lcddelay = millis() + 1500;
                       sprintf(report, "The volume is now %d (unmuted).", volume);
                       Serial.println(report);
                    }
                    else
                    {
                       audioplayer.setVolume(0);
                       // don't write volume 0 to NVRAM
                       lcd_output("     Audio", "     Muted");
                       lcddelay = millis() + 1500;
                       sprintf(report, "Audio is now muted (press [M] again to unmute, or change volume).");
                       Serial.println(report);
                    }
                    break;

         case 'N' : // Next station
         case 'n' : change_station(NEXT);
                    break;

         case 'P' : // Previous station
         case 'p' : change_station(PREV);
                    break;

         case 'Q' : // Quieter
         case 'q' : change_volume(VOLDOWN);
                    break;

         case 'R' : // Reset
         case 'r' : lcd_output("RESTARTING!", "(via serial)");
                    Serial.println("RESTARTING!");
                    delay(500);
                    ESP.restart();
                    break;

         case 'S' : // Setup
         case 's' : presetup();
                    break;

         case 'W' : // List WiFi APs (undocumented)
         case 'w' : wifi_listaps();
                    break;
      }
   }

   return;
}



// *******************************
// *                             *
// * SCAN AND LIST ACCESS POINTS *
// *                             *
// *******************************

int  wifi_listaps(void)
{
   // List discovered access points

   int  scanfound = 0;
   int  recurse = 0;
   char report[64];
   char apname[64];

   Serial.println("Scanning for available access points...");

   scanfound = WiFi.scanNetworks();

   Serial.println("");
   Serial.print("Discovered ");
   Serial.print(scanfound);
   if ( scanfound == 1)
      Serial.println(" WiFi access point.");  // messages line "Found 1 things" bug the hell out of me
   else
      Serial.println(" WiFi access points."); // it isn't THAT hard to sort out singular and plural!

   if ( scanfound > 0 )
   {
      Serial.println(" #  Name                                  Sig  Ch  Enc?");
      Serial.println("-------------------------------------------------------");

      for ( recurse = 0; recurse < scanfound; recurse++)
      {
         strcpy(apname, WiFi.SSID(recurse).c_str());
         apname[36] = '\0'; // hard clip it if longer
         sprintf(report, "%2d  %-36s  %2d %2d   %c",
                 (recurse + 1), apname,
                 WiFi.RSSI(recurse), WiFi.channel(recurse),
                 ((WiFi.encryptionType(recurse) == WIFI_AUTH_OPEN) ? ' ' : '*') );
         Serial.println(report);
      }
      Serial.println("");
   }

   return scanfound;
}



// ******************************
// *                            *
// * SOME SERIAL INPUT HANDLERS *
// *                            *
// ******************************

void wait_enter(void)
{
   // Wait for Enter to be pressed
   // Accepts either LF or CR here.

   int  inbyte = 0;

   Serial.print("Press [Enter] to continue...");
   while (1)
   {
      // Await some sort of serial input
      if ( Serial.available() > 0 )
      {
         inbyte = Serial.read();

         if ( (inbyte == 10) || (inbyte == 13) )
         {
            // If there's a byte available, swallow it.
            // At this point, it's probably the LF of a CRLF pair.
            if ( Serial.available() > 0 )
               inbyte = Serial.read();

            Serial.println("");
            Serial.println("");
            return;
         }
      } // end of if data available block
   } // end of never-ending loop

   // we never come here
}



int  wait_yesno(void)
{
   // Wait for Y or N to be pressed.
   // Returns TRUE if Yes, or FALSE if No.
   // Assumes question has already been asked.
   int  inbyte = 0;

   while (1)
   {
      // Await some sort of serial input
      if ( Serial.available() > 0 )
      {
         inbyte = Serial.read();

         if ( (inbyte == 'Y') || (inbyte == 'y') )
         {
            // Yes
            Serial.println("Yes");
            return true;
         }

         if ( (inbyte == 'N') || (inbyte == 'n') )
         {
            // No
            Serial.println("No");
            return false;
         }

      } // end of if data available block
   } // end of never-ending loop

   // we never come here
}



char *readline(char *orig, int maxlen)
{
   // Set up a simple text entry in the form:
   //   [<string here>       ]
   // where user can enter text, up to maxlen-1 bytes (final byte is termination)
   // until Enter pressed. Supports using Delete to erase content.
   // If "orig" is specified, this is used as the initial text which can be edited.

   static char instr[128]; // 128 bytes max, static so it can be returned
   int  inbyte = 0;
   int  posn = 0;
   int  slen = 0;
   char del[2] = { 8, 0 };

   // Set original text
   strcpy(instr, orig);

   // Set lengths
   if ( maxlen > 128 )
      maxlen = 128;
   slen = maxlen - 1;

   //  Print editor shell (assume it is all one line, so max length is practically ~72 chars max)
   Serial.print("[");
   for ( posn = 0; posn < maxlen; posn++ )
      Serial.print(" ");
   Serial.print("]");
   for ( posn = 0; posn <= maxlen; posn++ )
      Serial.print(del); // terminal should NOT erase characters on a backspace

   // Set initial text
   Serial.print(instr);
   posn = strlen(instr);

   while (1)
   {
      // Await some sort of serial input
      if ( Serial.available() > 0 )
      {
         inbyte = Serial.read();

         switch (inbyte)
         {
            case  10 : // LF, swallow these
                       break;

            case  13 : // CR, accept this as the input
                       return instr;

            case   8 : // Backspace
            case 127 : // Delete
                       if ( posn > 0)
                       {
                          instr[--posn] = '\0'; // delete character from buffer
                          Serial.print(del);    // and from the screen
                          Serial.print(" ");
                          Serial.print(del);
                       }
                       break;

            default  : // Add the character
                       if ( ( inbyte >= 32) && ( posn < slen ) )
                       {
                          char out[2] = { 0, 0 };
                          instr[posn++] = inbyte; // update array
                          instr[posn] = '\0';     // ensure it is always terminated
                          out[0] = inbyte;
                          Serial.print(out);      // update screen
                       }
                       break;
         } // end of byte read switch block
      } // end if is data available block
   } // end of never-ending block

   // we never come here
}



// **************************
// *                        *
// * SAVED STATION ROUTINES *
// *                        *
// **************************

int  savedstation_count(void)
{
   // Count how many stations are defined.

   int  recurse = 0;
   int  count = 0;
   char url[128] = "";
   char idx[16] = "";

   for ( recurse = 0; recurse < stationcount; recurse++ )
   {
      sprintf(idx, "stn%durl", recurse);
      strncpy(url, settings.getString(idx, "").c_str(), 127);
      url[127] = '\0';
      if ( strlen(url) > 12 )
         count++; // Why 12? Because it must be at least "http://x.yz/"
   }

   return count;
}



void savedstation_loadall(void)
{
   // Load the defined stations

   int  recurse = 0;
   char idx[16] = "";

   for ( recurse = 0; recurse < stationcount; recurse++ )
   {
      sprintf(idx, "stn%dname", recurse);
      strncpy(station[recurse].name, settings.getString(idx, "").c_str(), 63);
      station[recurse].name[63] = '\0';
      sprintf(idx, "stn%durl", recurse);
      strncpy(station[recurse].url, settings.getString(idx, "").c_str(), 127);
      station[recurse].url[127] = '\0';
   }

   return;
}



void savedstation_save(int which)
{
   // Save the specified station to NVRAM.

   char idx[16] = "";
   String tosave;

   // Save the station name
   sprintf(idx, "stn%dname", which);
   tosave = station[which].name; // make it a String
   settings.putString(idx, tosave);

   // Save the station name
   sprintf(idx, "stn%durl", which);
   tosave = station[which].url; // make it a String
   settings.putString(idx, tosave);

   return;
}



void savedstation_saveall(void)
{
   // Save all stations to NVRAM.

   int  recurse = 0;

   for ( recurse = 0; recurse < stationcount; recurse++ )
      savedstation_save(recurse);

   return;
}



// *************************
// *                       *
// * SYSTEM SETUP ROUTINES *
// *                       *
// *************************

void presetup(void)
{
   // Called at init if we don't have settings.
   //
   // The body of the function is duplicated, because we have two different messages on the LCD,
   // that we switch between every two seconds.

   int  inbyte = 0;

   while (1)
   {
      lcd_output("Set me up using", "a serial cable.");
      Serial.println("Press [Enter] to begin setup.");
      delay(2000); // two seconds

      if ( Serial.available() > 0 )
      {
         inbyte = Serial.read();
         if ( (inbyte == 10) || (inbyte == 13) )
         {
            // We have either CR or LF, so go on to do the setup.
            syssetup();
            return;
         }
      }

      // 115200bps, 8 bit, no parity, 1 stop bit, no flow control
      lcd_output("115200,8N1,NoFlw", "[Enter] to begin");
      delay(2000); // two seconds

      if ( Serial.available() > 0 )
      {
         inbyte = Serial.read();
         if ( (inbyte == 10) || (inbyte == 13) )
            return;
      }
   } // end of while (1)

   // we never come here
}



void syssetup(void)
{
   // Handle system configuration

   int  recurse = 0;
   int  inbyte = 0;
   char report[128] = "";

   // In case we come here from a functioning system, say we're
   // not connected as our time in here will mess up the stream,
   // so let's make it reconnect on the way back out.
   connected = 0;

   lcd_output("     Serial", "   Setup Mode");

   // Never-ending loop waiting for something to happen.
   while (1)
   {
      Serial.println("");
      Serial.println("SYSTEM SETUP");
      Serial.println("============");
      Serial.println("");

      if ( strlen(wifissid) > 0 )
      {
         // Configured, so report name and password
         Serial.print("WiFi AP : \"");
         Serial.print(wifissid);
         Serial.println("\"");
         Serial.print("Password: \"");
         Serial.print(wifipass);
         Serial.println("\"");
      }
      else
      {
         // Not configured, so say so.
         Serial.println("WiFi AP : <not configured>");
      }
      Serial.println("");

      Serial.println("Stations:");
      for (recurse = 0; recurse < 10; recurse++)
      {
         if ( strlen( station[recurse].url) == 0 )
            sprintf(report, " %d: undefined", recurse);
         else
            sprintf(report, " %d: %s (%s)", recurse, station[recurse].name, station[recurse].url);
         Serial.println(report);
      }
      Serial.println("");

      Serial.println("Note:  It is NOT recommended to use the Arduino IDE Serial Monitor,");
      Serial.println("       use a terminal that supports Backspace and Enter = CR or CRLF.");
      Serial.println("");
      Serial.println("Press: [W] to set up WiFi, [0]-[9] to set up the numbered station,");
      Serial.println("       or: [E]rase settings, [Q]uit setup, or [R]estart the radio.");
      Serial.println("");
      Serial.print(">"); // look ma, we have a prompt!

      inbyte = 0;
      do
      {
         // Await some sort of serial input
         if ( Serial.available() > 0 )
         {
            inbyte = Serial.read();

            switch (inbyte)
            {
               case  10 : // LF - we swallow these to stop the above text happening twice if user has CRLF termination
                          inbyte = 0;
                          break;

               case  13 : // CR, we'll output a newline prior to repeating the above text
                          Serial.println("");
                          break;

               case '0' : // Edit a station
               case '1' :
               case '2' :
               case '3' :
               case '4' :
               case '5' :
               case '6' :
               case '7' :
               case '8' :
               case '9' : Serial.println("Edit station");
                          syssetup_editstation( inbyte - '0' ); // convert ASCII to number
                          break;

               case 'E' : // Erase settings
               case 'e' : Serial.println("Erase settings");
                          syssetup_erase();
                          break;

               case 'Q' : // Quit - exit this function back to whence we came
               case 'q' : Serial.println("Quit setup");
                          // Ensure currentstation still points to a valid station
                          while ( strlen(station[currentstation].url) == 0 )
                          {
                             currentstation++;
                             if ( currentstation == stationcount )
                                currentstation = 0;
                          }
                          return;

               case 'R' : // Restart - exactly what it says on the tin
               case 'r' : Serial.println("Restart");
                          ESP.restart();
                          break;

               case 'W' : // WiFi settings
               case 'w' : Serial.println("WiFi settings");
                          syssetup_wifi();
                          break;
            } // end of inbyte switch
         } // end if is data available block
      } while ( inbyte == 0 ); // end of await input block
   } // end of config function body

   // never exits to here
}



void syssetup_wifi(void)
{
   // Set up the WiFi access point

   char ourssid[64] = "";
   char ourpass[64] = "";
   int  recurse = 0;
   int  result = 0;
   int  inbyte = 0;
   int  timeout = 0;
   String tosave;

   // Scan for access points, but only do it a maximum of five times.
   do
   {
      if ( recurse > 0 )
      {
         sprintf(ourssid, "No access points found, rescanning (try %d of 5)", (recurse + 1));
         Serial.println(ourssid);
      }
      result = wifi_listaps();
      recurse++;
   } while ( ( result == 0 ) && ( recurse < 5) );

   // If nothing found, bomb out.
   if ( result == 0 )
   {
      Serial.println("No access points found, unable to continue setting up WiFi.");
      Serial.println("");
      wait_enter(); // user must press Enter, so we know they've read the message.
      return;
   }

   // Choose an access point name
   sprintf(ourssid, "Which access point (1-%d, or X to abort) ? ", result);
   Serial.print(ourssid);
   recurse = 0;
   do
   {
      if ( Serial.available() > 0 )
      {
         inbyte = Serial.read();

         if ( ( inbyte == 'X' ) || ( inbyte == 'x' ) )
         {
            Serial.println("Aborted");
            return;
         }

         if ( ( inbyte > '0' ) && ( inbyte <= '9' ) )
         {
            recurse = inbyte - '0'; // ASCII to number
            if ( recurse > result)  // out of range
               recurse = 0;
         }
      }
   } while ( recurse == 0 );
   strcpy(ourssid, WiFi.SSID(recurse - 1).c_str());

   Serial.println("");
   Serial.print("Selected access point \"");
   Serial.print(ourssid);
   Serial.println("\".");
   Serial.println("");

   // Get the WiFi password
   Serial.print("Password: ");
   strcpy(ourpass, readline("", 64));

   // Check it
   Serial.println("");
   Serial.print("Verifying...");

   // If we're connected to a network, disconnect
   if ( WiFi.status() == WL_CONNECTED )
      WiFi.disconnect();

   // Now try to connect
   WiFi.begin(ourssid, ourpass);

   timeout = 0;

   while ( (WiFi.status() != WL_CONNECTED) && (timeout < 10) )
   {
      delay(350);
      Serial.print(".");
      timeout += 1;
   }

   if (WiFi.status() == WL_CONNECTED)
   {
      // Successful!
      Serial.println("connected.");
      Serial.println("Your radio will connect to this access point from now on.");
      strcpy(wifissid, ourssid);
      strcpy(wifipass, ourpass);
   }
   else
   {
      Serial.println("failed.");
      Serial.println("Saving the values given, but if connection keeps");
      Serial.println("failing, you should check the password is correct.");
      WiFi.disconnect(); // ensure we're not stuck
   }
   Serial.println("");

   // Always save the entered data, because the user won't appreciate
   // it if a glitch means we throw away a long WPA password that was
   // a pain in the ass to enter...

   tosave = ourssid; // SSID as a String
   settings.putString("ssid", tosave);

   tosave = ourpass; // Pass as a String
   settings.putString("pass", tosave);

   // Wait on Enter so we know they've read the above message
   wait_enter();

   return;
}



void syssetup_editstation(int which)
{
   // Edit the specified station

   char tmp[64] = "";
   int  recurse = 0;
   int  result = 0;
   int  inbyte = 0;
   int  timeout = 0;
   String tosave;

   sprintf(tmp, "Editing station %d", which);
   Serial.println(tmp);
   Serial.println("");

   Serial.println("Please enter the station name.");
   Serial.print("  ");
   strcpy(station[which].name, readline(station[which].name, 63));
   Serial.println("");

   Serial.println("");
   Serial.println("Please enter the station URL. This should be a full URL in the");
   Serial.println("form http://domain.tld/ or http://domain.tld:port/, with or");
   Serial.println("without a path. Please note that the stream must be MP3 (not OGG");
   Serial.println("or AAC) and that https is not supported.");
   Serial.print("  ");
   strcpy(station[which].url, readline(station[which].url, 74)); // don't let it go over the screen width
   Serial.println("");
   
   Serial.println("");
   sprintf(tmp, "Thank you, saving station %d as:", which);
   Serial.println(tmp);
   Serial.print("Name: ");
   Serial.println(station[which].name);
   Serial.print("URL : ");
   Serial.println(station[which].url);
   Serial.println("");

   savedstation_save(which);

   // Wait on Enter so we know they've read the above message
   wait_enter();

   return;
}



void syssetup_erase(void)
{
   // Erase all saved settings.
   int  recurse = 0;

   // This is a descructive operation, so confirm it.
   Serial.println("");
   Serial.print("Erase WiFi and channel settings - sure? (y/n) ");
   if ( wait_yesno() == false )
      return; // decided not to

   // Nuke the NVRAM
   Serial.println("Erasing...");
   settings.clear();

   // Now clear the variables that we have
   strcpy(wifissid, "");
   strcpy(wifipass, "");
   volume = INITVOLUME;
   currentstation = 0;
   for ( recurse = 0; recurse < stationcount; recurse++ )
   {
      strcpy(station[recurse].name, "");
      strcpy(station[recurse].url, "");
   }

   // Note that if the system starts up with NO stations in NVRAM, it
   // will provide the default selection until configured otherwise.

   return;
}
